def unique_by(seq, idfun=None):
	"""Return a unique list, in order, with an optional property filter.

	>>> unique_by([1, 2, 2, 3])
	[1, 2, 3]
	>>> unique_by([{'prop': 1}, {'prop': 2}, {'prop': 2}, {'prop': 3}], lambda x: x.prop)
	[{'prop': 1}, {'prop': 2}, {'prop': 3}]
	"""
	# order preserving
	if idfun is None:
		def idfun(x):
			return x
	seen = []
	result = []
	for item in seq:
		marker = idfun(item)
		# in old Python versions:
		# if seen.has_key(marker)
		# but in new ones:
		if marker in seen:
			continue
		seen.append(marker)
		result.append(item)
	return result


def first_of(iterable, default=None):
	"""Return the first item of iterable argument,
	or None if the argument is not iterable"""
	if iterable:
		for item in iterable:
			return item
	return default


def red(x):
	print ""
	print '\033[1;31m@-->%r\033[1;m' % x
	print '\033[1;37m-------------------\033[1;m'



"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

from django.test import TestCase
from django.template import Template
from django.test.client import Client
import logging

from ticketing.models import Location

logger = logging.getLogger(__name__)

class MetaTagsTestCase(TestCase):
	#for quick tests
	def _pre_setup(self):
		pass

	def setUp(self):
		self.t1 = Template("\n\
			{% load meta %}\n\
			{% meta %}\n\
				<title>{% block pagetitle %}Blue Man Group{% endblock %}</title>\n\
				<meta name='description' content='' />\n\
				<meta name='author' content='xxx' />\n\
			{% endmeta %}")

		self.t2 = Template("\n\
			{% extends t1 %}\n\
			{% load meta %}\n\
			{% meta %}\n\
				<meta name='test' content='lol' />\n\
				<meta name='author' content='yyy' />\n\
			{% endmeta %}")

	def test_view_for_meta(self):

		l = Location.objects.get(slug='las-vegas')
		c = Client()
		response = c.head(l.get_absolute_url())

		print "@-->response %r" % dir(response)

		print "@-->location %r" % l.get_absolute_url()

		print "@-->context %r" % response.request.get('PATH_INFO')
		#print "@-->request %r" % response.context.get('request')
		request = response.context.get('request')
		print "@-->path %r" % request.get_full_path()

		self.assertEqual(1, 0)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'SEOModelDefault'
        db.create_table('seo_cascade_seomodeldefault', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('meta', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('omit', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('model', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255)),
            ('priority', self.gf('django.db.models.fields.DecimalField')(default=0.5, null=True, max_digits=2, decimal_places=1, blank=True)),
        ))
        db.send_create_signal('seo_cascade', ['SEOModelDefault'])

        # Adding model 'SEOPageOverride'
        db.create_table('seo_cascade_seopageoverride', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('meta', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('omit', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('path', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('image', self.gf('django.db.models.fields.files.ImageField')(max_length=100, null=True, blank=True)),
        ))
        db.send_create_signal('seo_cascade', ['SEOPageOverride'])


    def backwards(self, orm):
        
        # Deleting model 'SEOModelDefault'
        db.delete_table('seo_cascade_seomodeldefault')

        # Deleting model 'SEOPageOverride'
        db.delete_table('seo_cascade_seopageoverride')


    models = {
        'seo_cascade.seomodeldefault': {
            'Meta': {'ordering': "['model']", 'object_name': 'SEOModelDefault'},
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'meta': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'omit': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'priority': ('django.db.models.fields.DecimalField', [], {'default': '0.5', 'null': 'True', 'max_digits': '2', 'decimal_places': '1', 'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'})
        },
        'seo_cascade.seopageoverride': {
            'Meta': {'ordering': "['path']", 'object_name': 'SEOPageOverride'},
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.files.ImageField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'meta': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'omit': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'path': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['seo_cascade']



from django.core.exceptions import MiddlewareNotUsed
from django.contrib.sitemaps import GenericSitemap
from django.db import models
from django.core import urlresolvers
from django.conf.urls.defaults import url

from models import SEOModelDefault, SEOPageOverride
from util import first_of

class SEOMiddleware(object):
	def __init__(self):
		generate_sitemap()
		raise MiddlewareNotUsed('Sitemap generation complete.')

def generate_sitemap():
	valid_models = filter(lambda m: hasattr(m, 'get_absolute_url'), models.get_models())
	all_overrides = map(lambda x: x.path, SEOPageOverride.objects.filter(omit=True))
	sitemaps = {}

	for model in valid_models:
		model_str = "%s.%s" % (model.__module__, model.__name__)
		defaults = first_of(SEOModelDefault.objects.filter(model=model_str))
		# TODO: make this a setting
		priority = .5

		if defaults:
			if defaults.omit:
				continue

			if defaults.priority:
				priority = defaults.priority

		# get list of items that have not been omitted by the admin
		allowed_items = map(lambda i: i.pk, filter(lambda x: not x.get_absolute_url() in all_overrides, model.objects.all()))
		queryset = model.objects.filter(pk__in=allowed_items)

		info_dict = {
			'queryset': queryset
		}

		sitemaps[model_str] = GenericSitemap(info_dict, priority=priority)

	resolver = urlresolvers.get_resolver(None)
	resolver.url_patterns.append(url(r'^sitemap\.xml$', 'django.contrib.sitemaps.views.sitemap', {'sitemaps': sitemaps}))



import ez_setup
ez_setup.use_setuptools()
from setuptools import setup, find_packages

setup(
		name                 = "django-seo-cascade",
		version              = "0.2",
		packages             = find_packages(),
		author               = "Lynn Dylan Hurley, Jack Shedd",
		author_email         = "lynn.dylan.hurley@gmail.com",
		description          = "Easy SEO management. Automatic sitemap + easy meta management.",
		url                  = "https://github.com/thisismess/django-seo-cascade",
		include_package_data = True,
		install_requires     = ['Django>=1.3.1', 'lxml>=2.3.3', 'ez_setup']
)



from django.db import models
from lxml import etree
from django.contrib.sites.models import Site
from django.conf import settings


MODEL_CHOICES = [("%s.%s" % (m.__module__, m.__name__), ("%s.%s" % (m.__module__, m.__name__))) for m in filter(lambda x: hasattr(x, 'get_absolute_url'), models.get_models())]


class SEOBase(models.Model):
	title       = models.CharField(max_length=255, blank=True, null=True)
	description = models.TextField(blank=True, null=True)
	meta        = models.TextField("Meta Override HTML", blank=True, null=True)
	omit        = models.BooleanField("Omit from sitemap", default=False)

	class Meta:
		abstract = True

	@property
	def title_tags(self):
		if not self.title:
			return u""

		tags = []
		title_tag = etree.Element("title")
		title_tag.text = self.title
		tags.append(title_tag)
		tags.append(etree.Element("meta", content=self.title, property="og:title"))
		return u'%s' % reduce(lambda x, y: x+y, map(lambda x: etree.tostring(x, pretty_print=True), tags))

	@property
	def description_tags(self):
		if not self.description:
			return u""

		tags = []
		tags.append(etree.Element("meta", content=self.description, property="description"))
		tags.append(etree.Element("meta", content=self.description, property="og:description"))
		return u"%s" % reduce(lambda x, y: x+y, map(lambda x: etree.tostring(x, pretty_print=True), tags))


class SEOModelDefault(SEOBase):
	model       = models.CharField(max_length=255, blank=False, unique=True, choices=MODEL_CHOICES)
	priority    = models.DecimalField(max_digits=2, decimal_places=1, default=0.5, null=True, blank=True)

	class Meta:
		ordering = ['model',]
		abstract = False

	def __unicode__(self):
		return u'%s' % self.model


class SEOPageOverride(SEOBase):
	path        = models.CharField("Absolute Path", max_length=255, blank=False, null=False)
	image       = models.ImageField(upload_to="meta", blank=True, null=True)

	class Meta:
		ordering = ['path',]
		abstract = False

	def __unicode__(self):
		return u'%s' % self.path

	@property
	def image_tags(self):
		current_site = Site.objects.get(id=settings.SITE_ID)
		if not self.image:
			return u""

		tags = []
		tags.append(etree.Element("meta", content="http://%s%s" % (current_site.domain, self.image.url), property="og:image"))
		tags.append(etree.Element("meta", content="%r" % self.image.width, property="og:image:width"))
		tags.append(etree.Element("meta", content="%r" % self.image.height, property="og:image:height"))
		return u'%s' % reduce(lambda x, y: x+y, map(lambda x: etree.tostring(x, pretty_print=True), tags))



from django.contrib import admin
from models import SEOPageOverride, SEOModelDefault

class SEOPageOverrideAdmin(admin.ModelAdmin):
	fieldsets = (
		(None, {
			'fields': ('path',)
		}),
		('Meta Content', {
			'fields': ('title', 'description', 'image',)
			}),
		('Sitemap', {
			'fields': ('omit', )
		}),
		('Advanced', {
			'classes': ('collapse closed',),
			'fields': ('meta',)
			}),
	)

class SEOModelDefaultAdmin(admin.ModelAdmin):
	fieldsets = (
		(None, {
			'fields': ('model', 'omit',)
		}),
		('Advanced', {
			'classes': ('collapse closed',),
			'fields': ('meta',)
		}),
	)

admin.site.register(SEOModelDefault, SEOModelDefaultAdmin)
admin.site.register(SEOPageOverride, SEOPageOverrideAdmin)



from django import template
from django.template import VariableNode, FilterExpression, Parser
from django.template.base import TemplateSyntaxError, StringOrigin, UNKNOWN_SOURCE
from django.template.loader_tags import BlockNode, BLOCK_CONTEXT_KEY
from django.utils.safestring import mark_safe

from lxml import etree

from ..util import unique_by, first_of
from ..models import SEOPageOverride

register = template.Library()

class MetaNode(BlockNode):
	def __init__(self, nodelist, parent=None):
		# prepend {{ block.super }} to meta tag
		p = Parser('')
		t = 'block.super'
		fe = FilterExpression(t, p)
		vn = VariableNode(fe)
		# i don't even know
		vn.source = (StringOrigin(UNKNOWN_SOURCE), ('a',{} ))
		nodelist.insert(0, vn)

		# build block tag with block name "meta"
		super(MetaNode, self).__init__("meta-cascade", nodelist, parent)

	def render(self, context):
		raw = super(MetaNode, self).render(context)

		request = context.get('request')
		path = request.get_full_path()

		admin_path_override = first_of(SEOPageOverride.objects.filter(path=path))

		if admin_path_override:
			raw += admin_path_override.title_tags
			raw += admin_path_override.description_tags
			raw += admin_path_override.image_tags
			raw += admin_path_override.meta

		root = etree.XML('<root>%s</root>' % raw)
		children = root.getchildren()
		children.reverse()
		children = unique_by(children, lambda x: [x.tag, x.get('name'), x.get('property')])
		children.reverse()

		if len(children):
			ret = reduce(lambda x, y: x+y, map(lambda z: (etree.tostring(z)), children))
		else:
			ret = ""

		return ret

	def super(self):
		if not hasattr(self, 'context'):
			return ''
		render_context = self.context.render_context
		if (BLOCK_CONTEXT_KEY in render_context and
				render_context[BLOCK_CONTEXT_KEY].get_block(self.name) is not None):
				return mark_safe(self.render(self.context))
		return ''


def do_meta(parser, token):
	"""
	Define a meta block that can be overridden by child templates.
	"""
	bits = token.contents.split()
	if len(bits) != 1:
		raise TemplateSyntaxError("'%s' tag takes no arguments" % bits[0])
	block_name = bits[0]
	# Keep track of the names of MetaNodes found in this template, so we can
	# check for duplication.
	try:
		if block_name in parser.__loaded_blocks:
			raise TemplateSyntaxError("'%s' tag appears more than once" % bits[0])
		parser.__loaded_blocks.append(block_name)
	except AttributeError: # parser.__loaded_blocks isn't a list yet
		parser.__loaded_blocks = [block_name]
	nodelist = parser.parse(('endmeta', 'endmeta %s' % block_name))

	parser.delete_first_token()
	return MetaNode(nodelist)


register.tag('meta', do_meta)
